/*
 * Concurrent Programming
 */

/** 
 * @version 1.0
 */

#include "CPAlgorithm.h"

/******************************************************************************
 ******************************* INCLUDE SECTION ******************************
 ******************************************************************************/

#include <cassert>

// STL (standard template library)
#include <string>
#include <iostream>
#include <vector>
#include <chrono>
#include <thread>
#include <functional>

// STB Image
#define STB_IMAGE_IMPLEMENTATION
#include <stb_image.h>
#define STB_IMAGE_WRITE_IMPLEMENTATION
#include <stb_image_write.h>

// OpenMP
#include <omp.h>

/******************************************************************************
 ****************************** NAMESPACE SECTION *****************************
 ******************************************************************************/

// Project
using namespace CpCore;

/******************************************************************************
 ************************* DEFINE AND CONSTANT SECTION ************************
 ******************************************************************************/

/******************************************************************************
 ***************************** TYPE DEFINITION ********************************
 ******************************************************************************/

/******************************************************************************
 ***************************** METHOD DEFINITION ******************************
 ******************************************************************************/

/******************************************************************************
 * Constructor
 *
 * it uses initialisation list
 ******************************************************************************/
CPAlgorithm::CPAlgorithm()
:	mWidth( 0 )
,	mHeight( 0 )
,	mNbChannels( 0 )
,	mInputFilename()
,	mInputData()
,	mInputFloatData()
,	mNbCppThreadsSupported( 0 )
,	mNbOpenMPThreadsSupported( 0 )
,	mMeanFilterSize( 0 )
,	mOutputData()
,	mThreshold( 0.f )
{
}

/******************************************************************************
 * Desstructor
 ******************************************************************************/
CPAlgorithm::~CPAlgorithm()
{
}

/******************************************************************************
 * Initialization
******************************************************************************/
void CPAlgorithm::initialize()
{
	// C++ threads
	mNbCppThreadsSupported = std::thread::hardware_concurrency();

	// OpenMP
	mNbOpenMPThreadsSupported = omp_get_max_threads();
}

/******************************************************************************
 * Convert an rgb color to a greyscale intensity
 *
 * @return the converted value
 ******************************************************************************/
unsigned int CPAlgorithm::getWidth() const
{
	return mWidth;
}

/******************************************************************************
 * Convert an rgb color to a greyscale intensity
 *
 * @return the converted value
 ******************************************************************************/
unsigned int CPAlgorithm::getHeight() const
{
	return mHeight;
}

/******************************************************************************
 * Convert an rgb color to a greyscale intensity
 *
 * @return the converted value
 ******************************************************************************/
unsigned int CPAlgorithm::getNbChannels() const
{
	return mNbChannels;
}

/******************************************************************************
 * Convert an rgb color to a greyscale intensity
 *
 * @return the converted value
 ******************************************************************************/
const std::string& CPAlgorithm::getInputFilename() const
{
	return mInputFilename;
}

float CPAlgorithm::getThreshold() const
{
	return mThreshold;
}
void CPAlgorithm::setThreshold( const float value )
{
	mThreshold = value;
}

unsigned int CPAlgorithm::getBlurRadius() const
{
	return mMeanFilterSize;
}
void CPAlgorithm::setBlurRadius( const unsigned int value )
{
	mMeanFilterSize = value;
}

/******************************************************************************
 * Returns the number of concurrent threads supported by the implementation.
 * The value should be considered only a hint.
 *
 * @return Number of concurrent threads supported. If the value is not well defined or not computable, returns 0.
 ******************************************************************************/
unsigned int CPAlgorithm::getNbConcurrentThreadsCpp() const
{
	return mNbCppThreadsSupported;
}
unsigned int CPAlgorithm::getNbConcurrentThreadsOpenMP() const
{
	return mNbOpenMPThreadsSupported;
}

/******************************************************************************
 * Get input data
 ******************************************************************************/
const std::vector< std::uint8_t >& CPAlgorithm::getInputData() const
{
	return mInputData;
}

/******************************************************************************
 * Read input image
 *
 * @return the converted value
 ******************************************************************************/
bool CPAlgorithm::readImage( const std::string& filename )
{
	// Input image parameters
	std::cout << "\nLoad image..." << std::endl;
	std::string inputFilename = filename;
	int width, height, nbChannels;
	//unsigned char* data = stbi_load( inputFilename.c_str(),
	//                                 &width, &height, &nbChannels, 0/*no conversion*/ );
	int req_comp = 4;
	unsigned char* data = stbi_load( inputFilename.c_str(),
	                                 &width, &height, &nbChannels, req_comp/*RGBA conversion*/ );
	// - check data
	if ( data == nullptr )
	{
		std::cout << "Error while loading image..." << std::endl;
		return false;
	}
	// - log info
	std::cout << "- image: (" << width << "x" << height << ") pixels, " << nbChannels << " channel(s)" << std::endl;
	
	// Update member data
	mWidth = width;
	mHeight = height;
	//mNbChannels = nbChannels;
	mNbChannels = ( req_comp == 0 ) ? nbChannels : req_comp;
	mInputFilename = filename;
		
	// Use C++ STL container instead (copy data)
	std::cout << "- use C++ STL container instead..." << std::endl;
	mInputData.clear();
	mInputData.resize( mWidth * mHeight * mNbChannels );
	int i = 0; // counter
	for ( auto& v : mInputData )
	{
		v = static_cast< std::uint8_t >( data[ i ] );
		i++;
	}
	mInputFloatData.clear();
	mInputFloatData.resize( mWidth * mHeight * mNbChannels );
	i = 0; // reset counter
	for ( auto& v : mInputFloatData )
	{
		v = static_cast< float >( data[ i ] );
		i++;
	}

	// Initialize output (sme size as input)
	mOutputData.clear();
	mOutputData.resize( mWidth * mHeight * mNbChannels, 0/*default value*/ );
	
	// Free STB memory
	stbi_image_free( data );
	
    return true;
}

/******************************************************************************
 * Convert to grey
 *
 * @return flag telling whether or not it succeeds
 ******************************************************************************/
bool CPAlgorithm::convertToGrey( CPType type, const int nbThreads )
{
	bool status = false;
	
	switch ( type )
	{
		case CPType::eSequential:
			status = convertToGreySequential();
			break;
			
		case CPType::eThread:
			status = convertToGreyThread( nbThreads );
			break;
			
		case CPType::eOpenMP:
			status = convertToGreyOpenMP( nbThreads );
			break;
			
		default:
			status = convertToGreySequential();
			break;
	}
	
	return status;
}

/******************************************************************************
 * Convert to grey
 *
 * @return flag telling whether or not it succeeds
 ******************************************************************************/
bool CPAlgorithm::convertToGreySequential()
{	
	auto start = std::chrono::system_clock::now();
	
	// Convert to grey
	const int nbPixels = mWidth * mHeight;
	for ( int i = 0; i < nbPixels; i++ )
	{
		// Retrieve input data
		const std::uint8_t r = mInputData[ mNbChannels * i ];
		const std::uint8_t g = mInputData[ mNbChannels * i + 1 ];
		const std::uint8_t b = mInputData[ mNbChannels * i + 2 ];
		
		// Apply algorithm and write output data
		const std::uint8_t value = static_cast< std::uint8_t >( convertToGrey( r, g, b ) );
		mOutputData[ mNbChannels * i ] = value;
		mOutputData[ mNbChannels * i + 1 ] = value;
		mOutputData[ mNbChannels * i + 2 ] = value;
	}
	
	auto end = std::chrono::system_clock::now();
	std::chrono::duration< double > elapsed_seconds = end - start;
    std::cout << "- elapsed time: " << elapsed_seconds.count() << "s\n";
	
	// EXPORT resulting image with STB
	const std::string outputFilename = "mountain_GREY_Sequential.jpg";
	bool status = exportImage( outputFilename );
	assert( status );

	return true;
}

/******************************************************************************
 * Convert to grey
 *
 * @return flag telling whether or not it succeeds
 ******************************************************************************/
bool CPAlgorithm::convertToGreyThread( const int nbThreads )
{
	const int nbPixels = mWidth * mHeight;
	
	// Pool of threads
	std::vector< std::thread > threads;
	for ( int i = 0; i < nbThreads; i++ )
	{
		threads.push_back( std::thread( &CPAlgorithm::convertToGreyThreadWork,
		                                this,
		                                i, nbThreads ) );
	}
	
	// Wait for threads to complete
	for ( auto& t : threads )
	{
		t.join();
	}

	// EXPORT resulting image with STB
	const std::string outputFilename = "mountain_GREY_CppThread.jpg";
	bool status = exportImage( outputFilename );
	assert( status );
	
	return true;
}

/******************************************************************************
 * Convert to grey
 *
 * @return flag telling whether or not it succeeds
 ******************************************************************************/
void CPAlgorithm::convertToGreyThreadWork( const int threadID, const int nbThreads )
{	
	// Convert to grey
	const int nbPixels = mWidth * mHeight;
	for ( int i = threadID; i < nbPixels; i += nbThreads )
	{
		// Retrieve input data
		const std::uint8_t r = mInputData[ mNbChannels * i ];
		const std::uint8_t g = mInputData[ mNbChannels * i + 1 ];
		const std::uint8_t b = mInputData[ mNbChannels * i + 2 ];
		
		// Apply algorithm and write output data
		const std::uint8_t value = static_cast< std::uint8_t >( convertToGrey( r, g, b ) );
		mOutputData[ mNbChannels * i ] = value;
		mOutputData[ mNbChannels * i + 1 ] = value;
		mOutputData[ mNbChannels * i + 2 ] = value;
	}
}

/******************************************************************************
 * Convert to grey
 *
 * @return flag telling whether or not it succeeds
 ******************************************************************************/
bool CPAlgorithm::convertToGreyOpenMP( const int nbThreads )
{
	// Set the number of threads
	omp_set_num_threads( nbThreads );
	
	// Convert to grey
	const int nbPixels = mWidth * mHeight;
	#pragma omp parallel for
	for ( int i = 0; i < nbPixels; i++ )
	{
		// Retrieve input data
		const std::uint8_t r = mInputData[ mNbChannels * i ];
		const std::uint8_t g = mInputData[ mNbChannels * i + 1 ];
		const std::uint8_t b = mInputData[ mNbChannels * i + 2 ];
		
		// Apply algorithm and write output data
		const std::uint8_t value = static_cast< std::uint8_t >( convertToGrey( r, g, b ) );
		mOutputData[ mNbChannels * i ] = value;
		mOutputData[ mNbChannels * i + 1 ] = value;
		mOutputData[ mNbChannels * i + 2 ] = value;
	}

	// EXPORT resulting image with STB
	const std::string outputFilename = "mountain_GREY_OpenMP.jpg";
	bool status = exportImage( outputFilename );
	assert( status );
	
	return true;
}

/******************************************************************************
 * Threshold
 *
 * @return flag telling whether or not it succeeds
 ******************************************************************************/
bool CPAlgorithm::threshold( CPType type, const int nbThreads )
{
	bool status = false;
	
	switch ( type )
	{
		case CPType::eSequential:
			status = thresholdSequential();
			break;
			
		case CPType::eThread:
			status = thresholdThread( nbThreads );
			break;
			
		case CPType::eOpenMP:
			status = thresholdOpenMP( nbThreads );
			break;
			
		default:
			status = thresholdSequential();
			break;
	}
	
	return status;
}

/******************************************************************************
 * Convert to grey
 *
 * @return flag telling whether or not it succeeds
 ******************************************************************************/
bool CPAlgorithm::thresholdSequential()
{	
	return true;
}

/******************************************************************************
 * Convert to grey
 *
 * @return flag telling whether or not it succeeds
 ******************************************************************************/
bool CPAlgorithm::thresholdThread( const int nbThreads )
{
	return true;
}

/******************************************************************************
 * Convert to grey
 *
 * @return flag telling whether or not it succeeds
 ******************************************************************************/
void CPAlgorithm::thresholdThreadWork( const int threadID, const int nbThreads )
{	
}

/******************************************************************************
 * Convert to grey
 *
 * @return flag telling whether or not it succeeds
 ******************************************************************************/
bool CPAlgorithm::thresholdOpenMP( const int nbThreads )
{	
	return true;
}

/******************************************************************************
 * Apply mean filter (blur)
 *
 * @return flag telling whether or not it succeeds
 ******************************************************************************/
bool CPAlgorithm::meanFilter( CPType type, const int nbThreads )
{
	bool status = false;
	
	switch ( type )
	{
		case CPType::eSequential:
			status = meanFilterSequential();
			break;
			
		case CPType::eThread:
			status = meanFilterThread( nbThreads);
			break;
			
		case CPType::eOpenMP:
			status = meanFilterOpenMP( nbThreads );
			break;
			
		default:
			status = meanFilterSequential();
			break;
	}
	
	return status;
}

/******************************************************************************
 * Half-size mask filter
 * - footprint is: (2M+1)x(2M+1)
 ******************************************************************************/
void CPAlgorithm::setMeanFilterSize( const unsigned int size )
{
	assert( size > 0 );
	mMeanFilterSize = size;
}

/******************************************************************************
 * Apply mean filter (blur)
 *
 * @return flag telling whether or not it succeeds
 ******************************************************************************/
bool CPAlgorithm::meanFilterSequential()
{
	return true;
}

/******************************************************************************
 * Apply mean filter (blur)
 *
 * @return flag telling whether or not it succeeds
 ******************************************************************************/
bool CPAlgorithm::meanFilterThread( const int nbThreads )
{	
	return true;
}

/******************************************************************************
 * Convert to grey
 *
 * @return flag telling whether or not it succeeds
 ******************************************************************************/
void CPAlgorithm::meanFilterThreadWork( const int threadID, const int nbThreads )
{
}

/******************************************************************************
 * Apply mean filter (blur)
 *
 * @return flag telling whether or not it succeeds
 ******************************************************************************/
bool CPAlgorithm::meanFilterOpenMP( const int nbThreads )
{
	return true;
}

/******************************************************************************
 * Convert an rgb color to a greyscale intensity
 *
 * @return the converted value
 ******************************************************************************/
float CPAlgorithm::convertToGrey( const std::uint8_t r, const std::uint8_t g, const std::uint8_t b )
{
    return 0.21f * static_cast< float >( r ) + 0.71f * static_cast< float >( g ) + 0.07f * static_cast< float >( b );
}

/******************************************************************************
 * Convert an rgb color to a greyscale intensity
 *
 * @return the converted value
 ******************************************************************************/
float CPAlgorithm::convertToGrey( const float r, const float g, const float b )
{
    return 0.21f * r + 0.71f * g + 0.07f * b;
}

/******************************************************************************
 * Export output image
 *
 * @return flag telling whether or not it succeeds
 ******************************************************************************/
bool CPAlgorithm::exportImage( const std::string& filename )
{
	const int quality = 90; // value in [ 1 ; 100 ] ; 1=>low ; 100=>high ; 0=>default to 90 ; NOTE: higher increase image memory size
	const int status = stbi_write_jpg( filename.c_str(),
	                                   mWidth, mHeight, mNbChannels,
								       mOutputData.data(),
	                                   quality );
	// Check status
	if ( status == 0 )
	{
		// Handle error
		std::cout << "- error while writing image: " << filename << std::endl;
		return false;
	}

	return true;
}
