/*
 * Concurrent Programming
 */

/** 
 * @version 1.0
 */

#ifndef _CP_ALGORITHM_H_
#define _CP_ALGORITHM_H_

/******************************************************************************
 ******************************* INCLUDE SECTION ******************************
 ******************************************************************************/

#include <cstdint>

// STL (standard template library)
#include <string>
#include <vector>

/******************************************************************************
 ************************* DEFINE AND CONSTANT SECTION ************************
 ******************************************************************************/

/******************************************************************************
 ***************************** TYPE DEFINITION ********************************
 ******************************************************************************/

/******************************************************************************
 ******************************** CLASS USED **********************************
 ******************************************************************************/

/******************************************************************************
 ****************************** CLASS DEFINITION ******************************
 ******************************************************************************/

namespace CpCore
{

/** 
 * @class CPAlgorithm
 *
 * @brief The CPAlgorithm class provides ...
 *
 * ...
 */
class CPAlgorithm
{
	
	/**************************************************************************
	 ***************************** PUBLIC SECTION *****************************
	 **************************************************************************/

public:

	/******************************* INNER TYPES *******************************/
	
	/**
	 * Enumeration
	 */
	enum class CPType
	{
		eSequential,
		eThread,
		eOpenMP,
		eNbCPTypes
	};

	/******************************* ATTRIBUTES *******************************/

	/******************************** METHODS *********************************/

	/**
	 * Constructor.
	 */
	CPAlgorithm();
	
	/**
	 * Destructor.
	 */
	virtual ~CPAlgorithm();

	/**
	 * Initialization
	 */
	void initialize();
	
	/**
	 * Accessor(s)
	 */
	unsigned int getWidth() const;
	unsigned int getHeight() const;
	unsigned int getNbChannels() const;
	const std::string& getInputFilename() const;

	float getThreshold() const;
	void setThreshold( const float value );
	unsigned int getBlurRadius() const;
	void setBlurRadius( const unsigned int value );

	/**
	 * Returns the number of concurrent threads supported by the implementation.
	 * The value should be considered only a hint.
	 *
	 * @return Number of concurrent threads supported. If the value is not well defined or not computable, returns 0.
	 */
	unsigned int getNbConcurrentThreadsCpp() const;
	unsigned int getNbConcurrentThreadsOpenMP() const;
	
	/**
	 * Read input image
	 */
	bool readImage( const std::string& filename );
	
	/**
	 * Convert to grey
	 *
	 * @return flag telling whether or not it succeeds
	 */
	bool convertToGrey( CPType type, const int nbThreads = 0 );

	/**
	 * Threshold
	 *
	 * @return flag telling whether or not it succeeds
	 */
	bool threshold( CPType type, const int nbThreads = 0 );
	
	/**
	 * Apply mean filter (blur)
	 *
	 * @return flag telling whether or not it succeeds
	 */
	bool meanFilter( CPType type, const int nbThreads = 0 );
	
	/**
	 * Half-size mask filter
	 * - footprint is: (2M+1)x(2M+1)
	 */
	void setMeanFilterSize( const unsigned int size );

	/**
	 * Get input data
	 */
	const std::vector< std::uint8_t >& getInputData() const;

	bool exportImage( const std::string& filename );

	/**************************************************************************
	 **************************** PROTECTED SECTION ***************************
	 **************************************************************************/

protected:
	
	/******************************* ATTRIBUTES *******************************/
	
	/**
	 * Input image properties
	 */
	unsigned int mWidth;
	unsigned int mHeight;
	unsigned int mNbChannels;
	std::string mInputFilename;
	
	/**
	 * 1D Buffer to store input image data,
	 * either in std::uint8_t (values ranging from 0 to 255) or float
	 */
	std::vector< std::uint8_t > mInputData;
	std::vector< float > mInputFloatData;

	/**
	 * Output buffer
	 */
	std::vector< std::uint8_t > mOutputData;
	
	unsigned int mNbCppThreadsSupported;
	int mNbOpenMPThreadsSupported;

	/**
	 * Half-size mask filter
	 * - footprint is: (2M+1)x(2M+1)
	 */
	int mMeanFilterSize;

	/**
	 * Threshold
	 */
	float mThreshold;
	
	/******************************** METHODS *********************************/
	
	// Sequential algorithms
	bool convertToGreySequential();
	// Multi-threaded algorithms with STL C++ thread
	bool convertToGreyThread( const int nbThreads = 0 );
	void convertToGreyThreadWork( const int threadID, const int nbThreads );
	// Multi-threaded algorithms with OpenMP
	bool convertToGreyOpenMP( const int nbThreads = 0 );

	// Sequential algorithms
	bool thresholdSequential();
	// Multi-threaded algorithms with STL C++ thread
	bool thresholdThread( const int nbThreads = 0 );
	void thresholdThreadWork( const int threadID, const int nbThreads );
	// Multi-threaded algorithms with OpenMP
	bool thresholdOpenMP( const int nbThreads = 0 );

	// Sequential algorithms
	bool meanFilterSequential();
	// Multi-threaded algorithms with STL C++ thread
	bool meanFilterThread( const int nbThreads = 0 );
	void meanFilterThreadWork( const int threadID, const int nbThreads );
	// Multi-threaded algorithms with OpenMP
	bool meanFilterOpenMP( const int nbThreads = 0 );

	/**
	 * Convert an rgb color to a greyscale intensity
	 */
	static float convertToGrey( const std::uint8_t r, const std::uint8_t g, const std::uint8_t b );
	static float convertToGrey( const float r, const float g, const float b );

	/**************************************************************************
	 ***************************** PRIVATE SECTION ****************************
	 **************************************************************************/

private:
	
	/******************************* ATTRIBUTES *******************************/
	
	/******************************** METHODS *********************************/

}; // end of class

} // end of namespace

#endif
