// Dear ImGui: standalone example application for GLFW + OpenGL 3, using programmable pipeline
// (GLFW is a cross-platform general purpose library for handling windows, inputs, OpenGL/Vulkan/Metal graphics context creation, etc.)
// If you are new to Dear ImGui, read documentation from the docs/ folder + read the top of imgui.cpp.
// Read online: https://github.com/ocornut/imgui/tree/master/docs

// GL
#include <glad/glad.h>

// STL
#include <string>
#include <iostream>
#include <vector>
#include <fstream>
#include <algorithm>
#include <functional>
#include <numeric>
#include <iomanip>
#include <algorithm>
#include <chrono>

#include <cassert>

// STB Image
//#define STB_IMAGE_IMPLEMENTATION
#include <stb_image.h>
//#define STB_IMAGE_WRITE_IMPLEMENTATION
#include <stb_image_write.h>

// Project
#include "CPAlgorithm.h"

#include "imgui.h"
#include "imgui_impl_glfw.h"
#include "imgui_impl_opengl3.h"
#include <stdio.h>
#if defined(IMGUI_IMPL_OPENGL_ES2)
#include <GLES2/gl2.h>
#endif
#include <GLFW/glfw3.h> // Will drag system OpenGL headers

// [Win32] Our example includes a copy of glfw3.lib pre-compiled with VS2010 to maximize ease of testing and compatibility with old VS compilers.
// To link with VS2010-era libraries, VS2015+ requires linking with legacy_stdio_definitions.lib, which we do using this pragma.
// Your own project should not be affected, as you are likely to link with a newer binary of GLFW that is adequate for your version of Visual Studio.
#if defined(_MSC_VER) && (_MSC_VER >= 1900) && !defined(IMGUI_DISABLE_WIN32_FUNCTIONS)
#pragma comment(lib, "legacy_stdio_definitions")
#endif

static void glfw_error_callback(int error, const char* description)
{
    fprintf(stderr, "Glfw Error %d: %s\n", error, description);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
 * Create a shader
 */
bool createShader( GLuint& shaderProgram, const std::string& shaderSource );

bool createInputTexture( GLuint& texture, const unsigned int width, const unsigned int height, const std::vector< std::uint8_t >& data );
bool createOutputTexture( GLuint& texture, const unsigned int width, const unsigned int height );

/**
 * OpenGL debug management
 */
static void APIENTRY openglCallbackFunction(
    GLenum source,
    GLenum type,
    GLuint id,
    GLenum severity,
    GLsizei length,
    const GLchar* message,
    const void* userParam
);

/******************************************************************************
 * Main entry program
 *
 * @param argc Number of arguments
 * @param argv List of arguments
 *
 * @return flag telling whether or not it succeeds
 ******************************************************************************/
int main( int argc, char** argv )
{
    // Setup window
    glfwSetErrorCallback(glfw_error_callback);
    if (!glfwInit())
        return 1;

    // Decide GL+GLSL versions
    const char* glsl_version = "#version 460";
    glfwWindowHint( GLFW_CONTEXT_VERSION_MAJOR, 4 );
    glfwWindowHint( GLFW_CONTEXT_VERSION_MINOR, 6 );
    glfwWindowHint( GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE );  // 3.2+ only
    glfwWindowHint( GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE );            // 3.0+ only

    // Create window with graphics context
    GLFWwindow* window = glfwCreateWindow( 1280, 720, "S4x: Concurrent Programming", NULL, NULL );
    if (window == NULL)
        return 1;
    glfwMakeContextCurrent(window);
    glfwSwapInterval(1); // Enable vsync

    // Load OpenGL API/extensions
    // gladLoadGLLoader( (GLADloadproc)glfwGetProcAddress );
    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress)) {
        std::cout << "Failed to initialize OpenGL context" << std::endl;
        return -1;
    }

    // Setup Dear ImGui context
    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO& io = ImGui::GetIO(); (void)io;
    //io.ConfigFlags |= ImGuiConfigFlags_NavEnableKeyboard;     // Enable Keyboard Controls
    //io.ConfigFlags |= ImGuiConfigFlags_NavEnableGamepad;      // Enable Gamepad Controls

    // Setup Dear ImGui style
    ImGui::StyleColorsDark();
    //ImGui::StyleColorsClassic();

    // Setup Platform/Renderer backends
    ImGui_ImplGlfw_InitForOpenGL(window, true);
    ImGui_ImplOpenGL3_Init(glsl_version);

    // Load Fonts
    // - If no fonts are loaded, dear imgui will use the default font. You can also load multiple fonts and use ImGui::PushFont()/PopFont() to select them.
    // - AddFontFromFileTTF() will return the ImFont* so you can store it if you need to select the font among multiple.
    // - If the file cannot be loaded, the function will return NULL. Please handle those errors in your application (e.g. use an assertion, or display an error and quit).
    // - The fonts will be rasterized at a given size (w/ oversampling) and stored into a texture when calling ImFontAtlas::Build()/GetTexDataAsXXXX(), which ImGui_ImplXXXX_NewFrame below will call.
    // - Read 'docs/FONTS.md' for more instructions and details.
    // - Remember that in C/C++ if you want to include a backslash \ in a string literal you need to write a double backslash \\ !
    //io.Fonts->AddFontDefault();
    //io.Fonts->AddFontFromFileTTF("../../misc/fonts/Roboto-Medium.ttf", 16.0f);
    //io.Fonts->AddFontFromFileTTF("../../misc/fonts/Cousine-Regular.ttf", 15.0f);
    //io.Fonts->AddFontFromFileTTF("../../misc/fonts/DroidSans.ttf", 16.0f);
    //io.Fonts->AddFontFromFileTTF("../../misc/fonts/ProggyTiny.ttf", 10.0f);
    //ImFont* font = io.Fonts->AddFontFromFileTTF("c:\\Windows\\Fonts\\ArialUni.ttf", 18.0f, NULL, io.Fonts->GetGlyphRangesJapanese());
    //IM_ASSERT(font != NULL);

    // Our state
    bool show_demo_window = true;
    bool show_another_window = false;
    ImVec4 clear_color = ImVec4(0.45f, 0.55f, 0.60f, 1.00f);

    // OpenGL
//#if _DEBUG // Use a boolean in User Interface instead !!!!!!!!!!!!!!!!!!!!!
	// Enable graphics callback
	glEnable( GL_DEBUG_OUTPUT );
	glEnable( GL_DEBUG_OUTPUT_SYNCHRONOUS );
	glDebugMessageCallback( openglCallbackFunction, nullptr );
	glDebugMessageControl( GL_DONT_CARE, GL_DONT_CARE, GL_DONT_CARE, 0, NULL, true );
//#endif
    // Main object
    CpCore::CPAlgorithm* algorithm = nullptr;
    // Algorithm
    int nbRequestedThreadsCpp = 1;
    int nbRequestedThreadsOpenMP = 1;
    int algorithmImplementationType = 3; // GPU
    int algorithmType = 0;
    //float threshold = 0.0f;
    float threshold = 0.378f;
    int filterRadius = 0;
    // Shaders
    GLuint shaderProgram_threshold = 0;
    GLuint shaderProgram_convertToGrey = 0;
    GLuint shaderProgram_blur = 0;
    // Textures
    GLuint mOutputTexture;
    GLuint mInputTexture;
    // Device timer(s) [GPU]
    GLuint64 mGPUTime = 0;
    GLuint mGPUQueryTimeElapsed;
    glCreateQueries( GL_TIME_ELAPSED, 1, &mGPUQueryTimeElapsed );

    // Determine the OpenGL and GLSL versions
    const GLubyte* mVendor;
    const GLubyte* mRenderer;
    const GLubyte* mVersion;
    const GLubyte* mShadingLanguageVersion;
    GLint mMajorVersion;
    GLint mMinorVersion;
	mVendor = glGetString( GL_VENDOR );
	mRenderer = glGetString( GL_RENDERER );
	mVersion = glGetString( GL_VERSION );
	mShadingLanguageVersion = glGetString( GL_SHADING_LANGUAGE_VERSION );
	glGetIntegerv( GL_MAJOR_VERSION, &mMajorVersion );
	glGetIntegerv( GL_MINOR_VERSION, &mMinorVersion );
	
	// Determine the OpenGL and GLSL versions
	printf( "\n[Renderer Info]\n" );
	printf( "\tGL Vendor: %s\n", mVendor );
	printf( "\tGL Renderer: %s\n", mRenderer );
	printf( "\tGL Version (string): %s\n", mVersion );
	printf( "\tGL Version (integer): %d.%d\n", mMajorVersion, mMinorVersion );
	printf( "\tGLSL Version: %s\n", mShadingLanguageVersion );

    // Main loop
    while ( ! glfwWindowShouldClose( window ) )
    {
        // Poll and handle events (inputs, window resize, etc.)
        glfwPollEvents();

        // --------------------------------------------------------------------------------------------------------------------------------
        // CUSTUM CODE

        static bool isShaderInitialized = false;
        static int width = 0;
        static int height = 0;
        if ( ! isShaderInitialized )
        {

        const std::string computeShader_convertToGrey = "";

        const std::string computeShader_threshold = "";

        const std::string computeShader_blur = "";

        //////////////////////////////////////////////////
        //std::string v = "#version 150\n";
        //std::string c = ReadTextFile("common.glsl"); // read a string with the file contents
        //std::string s = ReadTextFile("mesh.vert");   // same here
        //GLchar const* files[] = { v.c_str(), c.c_str(), s.c_str() };
        //GLint lengths[] = { v.size(),  c.size(),  s.size() };
        //glShaderSource(id, 3, files, lengths);
        ////////////////////////////////////////////////////

         //////////////////////////////////////////////
    	// Initialize output texture
        int inputWidth = 0;
        int inputHeight = 0;
        {
            // Initialize all
            // - create main object
            algorithm = new CpCore::CPAlgorithm();
            algorithm->initialize();
            // Read input image
            //const std::string inputFilename = "mountain.jpg";
            const std::string inputFilename = "mountain_small.jpg";
            bool status = algorithm->readImage( inputFilename );
            assert( status );
            const std::vector< std::uint8_t >& inputData = algorithm->getInputData();
            inputWidth = algorithm->getWidth();
            inputHeight = algorithm->getHeight();

        }

        // Initialize output texture
        width = inputWidth;
        height = inputHeight;
       

        //----------------------------------------------------
        //////////////////////////////////////////////

        // UPDATE FLAG
        isShaderInitialized = true;
    }

        if ( algorithmImplementationType == 3 ) // GPU
        {
            // Select shader program
            GLuint shaderProgram = 0;
            switch ( algorithmType )
            {
                case 0:
                    break;

                case 1:
                    break;

                case 2:
                    break;

                default:
                    break;
            }

             // - uniforms
            switch ( algorithmType )
            {
                case 0: // shaderProgram_convertToGrey
                    break;

                case 1: // shaderProgram_threshold
                    break;

                case 2: // shaderProgram_blur
                    break;

                default:
                    break;
            }
        }
        // --------------------------------------------------------------------------------------------------------------------------------

        // Start the Dear ImGui frame
        ImGui_ImplOpenGL3_NewFrame();
        ImGui_ImplGlfw_NewFrame();
        ImGui::NewFrame();

        // 1. Show the big demo window (Most of the sample code is in ImGui::ShowDemoWindow()! You can browse its code to learn more about Dear ImGui!).
        /*if (show_demo_window)
            ImGui::ShowDemoWindow(&show_demo_window);*/

        // 2. Show a simple window that we create ourselves. We use a Begin/End pair to created a named window.
        {
            //static float f = 0.0f;
            //static int counter = 0;

            bool my_tool_active = true;
            ImGui::Begin( "S4x: Concurrent Programming", &my_tool_active, ImGuiWindowFlags_MenuBar ); // Create a window called "Hello, world!" and append into it.
            if ( ImGui::BeginMenuBar() )
            {
                if ( ImGui::BeginMenu( "File" ) )
                {
                    if ( ImGui::MenuItem( "Open..", "Ctrl+O" ) ) { /* Do stuff */ }
                    if ( ImGui::MenuItem( "Save", "Ctrl+S" ) )   { /* Do stuff */ }
                    if ( ImGui::MenuItem( "Close", "Ctrl+W" ) )  { my_tool_active = false; }
                    ImGui::EndMenu();
                }
                ImGui::EndMenuBar();
            }

            ImGui::Text( "Application average %.3f ms/frame (%.1f FPS)", 1000.0f / ImGui::GetIO().Framerate, ImGui::GetIO().Framerate );

            if ( ImGui::CollapsingHeader( "Application Settings" ) )
            {
                ImGui::ColorEdit3( "background color", (float*)&clear_color ); // Edit 3 floats representing a color
            }

            ImGui::Text( "Image: %dx%d pixels - %d channels", algorithm->getWidth(), algorithm->getHeight(), algorithm->getNbChannels() );

            if ( ImGui::CollapsingHeader( "Hardware / Device" ) )
            {
                // Determine the OpenGL and GLSL versions
                ImGui::Text( "[Renderer Info]" );
                ImGui::Separator();
                ImGui::Text( "GL Vendor: %s", mVendor );
                ImGui::Text( "GL Renderer: %s", mRenderer );
                ImGui::Text( "GL Version (string): %s", mVersion );
                ImGui::Text( "GL Version (integer): %d.%d", mMajorVersion, mMinorVersion );
                ImGui::Text( "GLSL Version: %s", mShadingLanguageVersion );
            }

            if ( ImGui::CollapsingHeader( "Implementation Type" ) )
            {
                ImGui::RadioButton( "sequential", &algorithmImplementationType, 0 ); /*ImGui::SameLine();*/
                ImGui::RadioButton( "C++ threads", &algorithmImplementationType, 1 );/* ImGui::SameLine();*/
                ImGui::RadioButton( "OpenMP threads", &algorithmImplementationType, 2 );
                ImGui::RadioButton( "GPU threads", &algorithmImplementationType, 3 );

                ImGui::Separator();
                ImGui::Text( "Parameters:" );

                 switch ( algorithmImplementationType )
                 {
                    case 0: // sequential
                        {
                            static std::chrono::duration< double > elapsed_seconds_sequential;
                            if ( ImGui::Button( "Apply##Sequential" ) )
                            {
                                std::cout << "\nsequential..." << std::endl;

                                auto start = std::chrono::system_clock::now();
                            
                                switch ( algorithmType )
                                {
                                    case 0:
                                        algorithm->convertToGrey( CpCore::CPAlgorithm::CPType::eSequential );
                                        break;

                                    case 1:
                                        algorithm->threshold( CpCore::CPAlgorithm::CPType::eSequential );
                                        break;

                                    case 2:
                                        algorithm->meanFilter( CpCore::CPAlgorithm::CPType::eSequential );
                                        break;

                                    default:
                                        algorithm->convertToGrey( CpCore::CPAlgorithm::CPType::eSequential );
                                        break;
                                }

                                auto end = std::chrono::system_clock::now();
                                elapsed_seconds_sequential = end - start;
                                std::cout << "- elapsed time: " << elapsed_seconds_sequential.count() << "s\n";
                            }
                            ImGui::Text( "Timing: %f s", static_cast< float >(elapsed_seconds_sequential.count() ) );
                        }
                        break;

                    case 1: // C++ threads
                        {
                            static std::chrono::duration< double > elapsed_seconds_cpp;
                            ImGui::Text( "Nb threads supported: %d", algorithm->getNbConcurrentThreadsCpp() );
                            ImGui::SliderInt( "nb requested threads", &nbRequestedThreadsCpp, 0, algorithm->getNbConcurrentThreadsCpp() );
                            if ( ImGui::Button( "Apply##ThreadC++" ) )
                            {
                                std::cout << "\nC++ threads..." << std::endl;

                                auto start = std::chrono::system_clock::now();
                            
                                switch ( algorithmType )
                                {
                                    case 0:
                                        algorithm->convertToGrey( CpCore::CPAlgorithm::CPType::eThread, nbRequestedThreadsCpp );
                                        break;

                                    case 1:
                                        algorithm->threshold( CpCore::CPAlgorithm::CPType::eThread, nbRequestedThreadsCpp );
                                        break;

                                    case 2:
                                        algorithm->meanFilter( CpCore::CPAlgorithm::CPType::eThread, nbRequestedThreadsCpp );
                                        break;

                                    default:
                                        algorithm->convertToGrey( CpCore::CPAlgorithm::CPType::eThread, nbRequestedThreadsCpp );
                                        break;
                                }
                            
                                auto end = std::chrono::system_clock::now();
                                elapsed_seconds_cpp = end - start;
                                std::cout << "- elapsed time: " << elapsed_seconds_cpp.count() << "s\n";
                            }
                            ImGui::Text("Timing: %f s", static_cast<float>(elapsed_seconds_cpp.count()));
                        }
                        break;

                    case 2: // OpenMP threads
                        {
                            static std::chrono::duration< double > elapsed_seconds_OpenMP;
                            ImGui::Text( "Nb threads supported: %d", algorithm->getNbConcurrentThreadsOpenMP() );
                            ImGui::SliderInt( "nb requested threads", &nbRequestedThreadsOpenMP, 0, algorithm->getNbConcurrentThreadsOpenMP() );
                            if ( ImGui::Button( "Apply##OpenMP" ) )
                            {
                                std::cout << "\nOpenMP..." << std::endl;
                            
                                auto start = std::chrono::system_clock::now();

                                switch ( algorithmType )
                                {
                                    case 0:
                                        algorithm->convertToGrey( CpCore::CPAlgorithm::CPType::eOpenMP, nbRequestedThreadsOpenMP );
                                        break;

                                    case 1:
                                        algorithm->threshold( CpCore::CPAlgorithm::CPType::eOpenMP, nbRequestedThreadsOpenMP );
                                        break;

                                    case 2:
                                        algorithm->meanFilter( CpCore::CPAlgorithm::CPType::eOpenMP, nbRequestedThreadsOpenMP );
                                        break;

                                    default:
                                        algorithm->convertToGrey( CpCore::CPAlgorithm::CPType::eOpenMP, nbRequestedThreadsOpenMP );
                                        break;
                                }

                                auto end = std::chrono::system_clock::now();
                                elapsed_seconds_OpenMP = end - start;
                                std::cout << "- elapsed time: " << elapsed_seconds_OpenMP.count() << "s\n";
                            }
                            ImGui::Text("Timing: %f s", static_cast<float>(elapsed_seconds_OpenMP.count()));
                        }
                        break;

                    case 3: // GPU threads
                        {
                            ImGui::Text( "Blocks grid: %dx%dx%d", algorithm->getWidth() / 32, algorithm->getHeight() / 32, 1 );
                            ImGui::Text( "Threads block: %dx%dx%d", 32, 32, 1 );
                            ImGui::Text( "Timing: %f ms", static_cast< float >( mGPUTime * 1.e-6 ) );
                        }
                        break;

                    default:
                        break;
                 }
            }

            if ( ImGui::CollapsingHeader( "Filter Type" ) )
            {
                ImGui::RadioButton( "convert to grey", &algorithmType, 0 ); /*ImGui::SameLine();*/
                ImGui::RadioButton( "threshold", &algorithmType, 1 );/* ImGui::SameLine();*/
                ImGui::RadioButton( "blur", &algorithmType, 2 );
                ImGui::RadioButton( "procedural generation", &algorithmType, 3 );

                ImGui::Separator();
                ImGui::Text( "Parameters:" );

                 switch ( algorithmType )
                 {
                    case 0: // convert to grey
                        break;

                    case 1: // threshold
                        if ( ImGui::SliderFloat( "threshold", &threshold, 0.f, 1.f ) )
                        {
                            algorithm->setThreshold( threshold );
                        }
                        break;

                    case 2: // blur
                        if ( ImGui::SliderInt( "blur radius", &filterRadius, 0, 16 ) )
                        {
                            algorithm->setMeanFilterSize( filterRadius );
                        }
                        break;

                    case 3: // procedural generation
                        break;

                    default:
                        break;
                 }
            }
          
            //ImGui::Checkbox("Demo Window", &show_demo_window);      // Edit bools storing our window open/close state
            //ImGui::Checkbox("Another Window", &show_another_window);

            ImGui::End();
        }

        bool show_input_image_window = true;
        bool show_output_image_window = true;
        if ( show_input_image_window )
        {
            ImGui::Begin( "Input Image Window", &show_input_image_window );
            //ImGui::Image( (void*)(mInputTexture), ImVec2( static_cast< float >( algorithm->getWidth() ), static_cast< float >( algorithm->getHeight() ) ) );
            ImGui::End();
        }
        if ( show_output_image_window )
        {
            ImGui::Begin( "Output Image Window", &show_output_image_window );
          /*  ImGui::Image( (void*)(mOutputTexture),
                ImVec2( static_cast< float >( algorithm->getWidth() ),
                    static_cast< float >( algorithm->getHeight() ) ) );*/
            ImGui::End();
        }

        //// 3. Show another simple window.
        //if (show_another_window)
        //{
        //    ImGui::Begin("Another Window", &show_another_window);   // Pass a pointer to our bool variable (the window will have a closing button that will clear the bool when clicked)
        //    ImGui::Text("Hello from another window!");
        //    if (ImGui::Button("Close Me"))
        //        show_another_window = false;
        //    ImGui::End();
        //}

        // Rendering
        ImGui::Render();
        int display_w, display_h;
        glfwGetFramebufferSize( window, &display_w, &display_h );
        glViewport( 0, 0, display_w, display_h );
        glClearColor( clear_color.x * clear_color.w, clear_color.y * clear_color.w, clear_color.z * clear_color.w, clear_color.w );
        glClear( GL_COLOR_BUFFER_BIT );
        ImGui_ImplOpenGL3_RenderDrawData( ImGui::GetDrawData() );

        // Swap buffers
        glfwSwapBuffers( window );
    }

    // Cleanup project
    delete algorithm;
    algorithm = nullptr;
    // - shader program(s)
    glDeleteProgram( shaderProgram_convertToGrey );
    glDeleteProgram( shaderProgram_threshold );
    glDeleteProgram( shaderProgram_blur );
    // - textures
    glDeleteTextures( 1, &mInputTexture );
    glDeleteTextures( 1, &mOutputTexture );
    // - device timers
    glDeleteQueries( 1, &mGPUQueryTimeElapsed );

    // Cleanup ImGui
    ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplGlfw_Shutdown();
    ImGui::DestroyContext();

    // Cleanup GLFW
    glfwDestroyWindow( window );
    glfwTerminate();

    // Exit program
    return 0;
}

/******************************************************************************
 * Create a shader
 ******************************************************************************/
bool createShader( GLuint& shaderProgram, const std::string& shaderSource )
{
    // Create shader
	GLuint shader_handle = glCreateShader( GL_COMPUTE_SHADER );

    // Set shader source code
    const GLchar* compute_shader_with_version[ 1 ] = { shaderSource.c_str() };
    glShaderSource( shader_handle, 1, compute_shader_with_version, NULL );
	
    // Compile shader
    glCompileShader( shader_handle );
	
    // Check compilation status
	GLint compileStatus;
	glGetShaderiv( shader_handle, GL_COMPILE_STATUS, &compileStatus );
	if ( compileStatus == GL_FALSE )
	{
		GLint logInfoLength = 0;
		glGetShaderiv( shader_handle, GL_INFO_LOG_LENGTH, &logInfoLength );
		if ( logInfoLength > 0 )
		{
			// Return information log for shader object
			GLchar* infoLog = new GLchar[ logInfoLength ];
			GLsizei length = 0;
			glGetShaderInfoLog( shader_handle, logInfoLength, &length, infoLog );

			// LOG
			std::cout << "\naddShader() - compilation ERROR" << std::endl;
			std::cout << infoLog << std::endl;

			delete[] infoLog;
		}

		return false;
	}

    // Create shader program
	shaderProgram = glCreateProgram();

    // Link shader
	glAttachShader( shaderProgram, shader_handle );
	glLinkProgram( shaderProgram );

	// Check linking status
	GLint linkStatus = 0;
	glGetProgramiv( shaderProgram, GL_LINK_STATUS, &linkStatus );
	if ( linkStatus == GL_FALSE )
	{
		GLint logInfoLength = 0;
		glGetProgramiv( shaderProgram, GL_INFO_LOG_LENGTH, &logInfoLength );
		if ( logInfoLength > 0 )
		{
			// Return information log for program object
			GLchar* infoLog = new GLchar[ logInfoLength ];
			GLsizei length = 0;
			glGetProgramInfoLog( shaderProgram, logInfoLength, &length, infoLog );

			// LOG
			std::cout << "\naddShader() - link ERROR" << std::endl;
			std::cout << infoLog << std::endl;

			delete[] infoLog;
		}

		return false;
	}
	
    // Detach and delete shader
	glDetachShader( shaderProgram, shader_handle );
	glDeleteShader( shader_handle );

    return true;
}

/******************************************************************************
 * OpenGL error management
 ******************************************************************************/
static void APIENTRY openglCallbackFunction(
  GLenum source,
  GLenum type,
  GLuint id,
  GLenum severity,
  GLsizei length,
  const GLchar* message,
  const void* userParam
)
{
  /*(void)source; (void)type; (void)id; 
  (void)severity; (void)length; (void)userParam;
  
  fprintf( stderr, "%s\n", message );
  
  if ( severity == GL_DEBUG_SEVERITY_HIGH )
  {
    fprintf( stderr, "Aborting...\n" );
    abort();
  }*/

	// TEST
	if ( severity != GL_DEBUG_SEVERITY_HIGH )
	//if ( severity != GL_DEBUG_SEVERITY_MEDIUM )
	//if ( severity != GL_DEBUG_SEVERITY_LOW )
	{
		return;
	}

	std::cout << "------------opengl-debug-callback------------" << std::endl;
	std::cout << "message: "<< message << std::endl;
	std::cout << "type: ";
    switch (type) {
    case GL_DEBUG_TYPE_ERROR:
		std::cout << "ERROR";
        break;
    case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
		std::cout << "DEPRECATED_BEHAVIOR";
        break;
    case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
		std::cout << "UNDEFINED_BEHAVIOR";
        break;
    case GL_DEBUG_TYPE_PORTABILITY:
		std::cout << "PORTABILITY";
        break;
    case GL_DEBUG_TYPE_PERFORMANCE:
		std::cout << "PERFORMANCE";
        break;
    case GL_DEBUG_TYPE_OTHER:
		std::cout << "OTHER";
        break;
    }
	std::cout << std::endl;
 
	std::cout << "id: " << id << std::endl;
	std::cout << "severity: ";
    switch (severity){
    case GL_DEBUG_SEVERITY_LOW:
		std::cout << "LOW";
        break;
    case GL_DEBUG_SEVERITY_MEDIUM:
		std::cout << "MEDIUM";
        break;
    case GL_DEBUG_SEVERITY_HIGH:
		std::cout << "HIGH";
        break;
    }
	std::cout << std::endl;
}

/******************************************************************************
 * Create input texture
 ******************************************************************************/
bool createInputTexture( GLuint& texture, const unsigned int width, const unsigned int height, const std::vector< std::uint8_t >& data )
{
    // Create texture
    glCreateTextures( GL_TEXTURE_2D, 1, &texture );
    // - texture parameters
    glTextureParameteri( texture, GL_TEXTURE_BASE_LEVEL, 0 );
    glTextureParameteri( texture, GL_TEXTURE_MAX_LEVEL, 0 );
    // - allocate memory
    const GLsizei levels = 1;
    const GLenum internalFormat = GL_RGBA8;
    glTextureStorage2D( texture, levels, internalFormat, width, height );
    // - fill data
    const GLenum format = GL_RGBA;
    const GLenum type = GL_UNSIGNED_BYTE;
    glTextureSubImage2D( texture, 0/*level*/, 0/*xoffset*/, 0/*yoffset*/, width, height, format, type, data.data() );

    return true;
}

/******************************************************************************
 * Create output texture
 ******************************************************************************/
bool createOutputTexture( GLuint& texture, const unsigned int width, const unsigned int height )
{
    // Create texture
    glCreateTextures( GL_TEXTURE_2D, 1, &texture );
    // - texture parameters
    glTextureParameteri( texture, GL_TEXTURE_BASE_LEVEL, 0 );
    glTextureParameteri( texture, GL_TEXTURE_MAX_LEVEL, 0 );
    // - allocate memory
    const GLsizei levels = 1;
    const GLenum internalFormat = GL_RGBA8;
    glTextureStorage2D( texture, levels, internalFormat, width, height );
    // - fill data
    const GLenum format = GL_RGBA;
    const GLenum type = GL_UNSIGNED_BYTE;
    std::uint32_t initValue = 0; // 4 * 8 = 32
    glClearTexImage( texture, /*level*/0, format, type, &initValue );

    return true;
}
